SpinToWin = LibStub("AceAddon-3.0"):NewAddon("SpinToWin", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SpinToWin")

--==== set upvalues ====--
-- GLOBALS: LibStub, LOOT_ITEM_PUSHED_SELF
local _G, SpinToWin, pairs, rawset, select, type, CreateFrame, strformat, mod, floor, max =
      _G, SpinToWin, pairs, rawset, select, type, CreateFrame, string.format, math.fmod, math.floor, math.max


SpinToWin.DL = {
	[0] = "silent",
	"error",
	"standard",
	"test",
	"config",
	"events",
	"spam",
}
local DL = SpinToWin.DL
do	-- add lookups for name to index and index as string to index to table  -- review - shouldn't it be index as string to name? check other addons
	for i=0, #DL do
		DL[ DL[i] ] = i
		DL[tostring(i)] = i
	end
end

local dbgprint = { print   = function(...) return SpinToWin:Print(...) end,
                   squelch = function(...) return nil end }
function SpinToWin:SetDebugLevel(iDL)
	for i = 0,     iDL do dbgprint[ DL[i] ] = dbgprint.print;   end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local cFlips = 0
local cMoney = 0
local tblFlipsSession = {}
local tblFlipsHistory
local dbglob

local tblCardLinks = setmetatable( {}, { 
		__index = function(t,k) 
			local _, link = GetItemInfo(k)
			rawset(t, k, link)
			return link
		end,
	})

local tblFlippableCards = { 113355, 113346, 113347, 113348, 113349 }
-- using metatable means we can get cost for Card of Omens, but not have it
-- included when iterating the table for all the set of possible flip results
local tblCardValue = setmetatable( {
		[113340] =        1;  -- "You will gain the sympathy of others."
		[113341] =     1000;  -- "You will see the world as it is."
		[113342] =     5000;  -- "You will overestimate your luck."
		[113343] =     5000;  -- "You will repair your armor soon."
		[113344] =    10000;  -- "You will worry, yet no good will come of it."
		[113345] =    10000;  -- "You will gain understanding of your enemies."
		[113346] =    50000;  -- "You will defy your destiny."
		[113347] =    50000;  -- "You will take chances."
		[113348] =   100000;  -- "You will consider the odds."
		[113349] =   200000;  -- "You will take an unexpected risk."
		[113350] =   500000;  -- "You will live a full life and die a glorious death."
		[113351] =  1000000;  -- "You will fall in love."
		[113352] = 10000000;  -- "You will visit a vendor soon."
		[113353] = 30000000;  -- "You will obtain something you desire."
		[113354] = 60000000;  -- "You will crow about your fortune."
	}, { __index = {
		[113355] =        0;  -- "Use: Flip the card to predict your fate."
	}, } )

local idCurrentCard = 113355
local cCost = tblCardValue[idCurrentCard]

local strItemPush
local function MakePatternFromFormat(strformat)
	local split_rules = {
		{ sub = "%1",   find = "(%%%%)" },		-- %% catch and deal with escaped %, make detection of other conditions a nightmare
		{ sub = "(.*)", find = "(%%%d+%$s)",  },		-- %11$s - not legal for format, some special blizz marker
		{ sub = "(.+)", find = "(%%[%-%+%s#0]*%d*%.?%d*[scdiuoxXfFeEgGaA])",   },    -- format specifiers, output cannot be empty string
		{ sub = "%%%1", find = "([%[%]%(%)%^%$%*%+%-%?%.])", }, -- escape out magic search characters
	}

	local function split(str, depth)
		if depth > #split_rules then
			return str
		end
		if str == "" then
			return str
		end

		local rule = split_rules[depth]
		local i, j, cap = strfind(str, rule.find)
		if not cap then
			return split(str, depth+1)
		elseif i == 1 then
			return gsub(cap, rule.find, rule.sub) .. split(strsub(str, j+1), depth)
		else
			return split(strsub(str, 1, i-1), depth+1) .. gsub(cap, rule.find, rule.sub) .. split(strsub(str, j+1), depth)
		end
	end

	strformat = strformat or ""
	return "^(" .. split(strformat, 1) .. ")$"
end

function SpinToWin:UNIT_SPELLCAST_SUCCEEDED(event, unitID, spell, rank, lineID, spellID, ...)
	-- dbgprint.test(event, unitID, spell, spellID)
	if 113355 == idCurrentCard and 166550 == spellID and "player" == unitID then
		-- we don't want to count reflips as flips because it will change our per warpaint averages
		-- dbgprint.test("card flipped")
		cFlips = cFlips + 1
		dbglob.totalFlips = dbglob.totalFlips + 1

		self.frm.SessionFlips:SetText(string.format("%i", cFlips))
		self.frm.HistoryFlips:SetText(string.format("%i", dbglob.totalFlips))
	end
end

function SpinToWin:ITEM_PUSH(event, bagID, icon)
	-- Getting a blood card by flipping or getting a card of omens 
	-- by crafting or from mail both use the same icon for this event
	-- Interface\\Icons\\INV_Misc_Ticket_Tarot_Wrath.blp
end


local function MyGetCoinTextureString(amount)
	local gold = floor(amount / 10000)
	local silver = floor(mod(amount, 10000) / 100)
	local copper = mod(amount, 100)
	-- print(amount, gold, silver, copper)
	return strformat("%01i|TInterface\\Icons\\INV_Misc_Coin_17:0|t %02i|TInterface\\Icons\\INV_Misc_Coin_18:0|t %02i|TInterface\\Icons\\INV_Misc_Coin_19:0|t   ", gold, silver, copper)
end


-- TODO - do we get a loot chat message from mail?
function SpinToWin:CHAT_MSG_LOOT(event, message)
	-- dbgprint.test(event, message)
	local strMatched, strItem = string.match(message, strItemPush)
	if strMatched and strItem then
		local itemId = string.match(strItem, "|Hitem:(%d+):")
		if itemId then
			itemId = tonumber(itemId)
			local value = tblCardValue[itemId]
			-- dbgprint.test(itemId, type(itemId), value)
			if value and value > 0 then
				tblFlipsSession[idCurrentCard][itemId] = tblFlipsSession[idCurrentCard][itemId] + 1
				tblFlipsHistory[idCurrentCard][itemId] = tblFlipsHistory[idCurrentCard][itemId] + 1

				local realvalue = value - cCost
				cMoney = cMoney + realvalue
				dbglob.totalMoney = dbglob.totalMoney + realvalue

				-- dbgprint.test("card", GetCoinTextureString(value), "session", GetCoinTextureString(cMoney), "lifetime", GetCoinTextureString(self.db.global.totalMoney))
				self.frm.SessionMoney:SetText(MyGetCoinTextureString(cMoney))
				self.frm.HistoryMoney:SetText(MyGetCoinTextureString(dbglob.totalMoney))
			end
		end
	end
end

local function btn_PreClick(self)
	local strCardCounts = ""
	for i=1, #tblFlippableCards do
		local id = tblFlippableCards[i]
		local count = GetItemCount(id)
		strCardCounts = strCardCounts .. tostring(count) .. "  "
	end
	dbgprint.test("btn_PreClick", strCardCounts)
end

function SpinToWin:UpdateButton()
end

local function frm_MouseWheel(self, delta)
end

local function frm_Click(self, button, down)
end

local function frm_Enter(self)
--	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
--	self:RegisterEvent("CHAT_MSG_LOOT")

--	SetOverrideBindingClick(self, true, "MOUSEWHEELUP",   "SpinToWinButton") -- card of omens
--	SetOverrideBindingClick(self, true, "MOUSEWHEELDOWN", "SpinToWinButton") -- card of omens

	-- local _, cardLink = GetItemInfo(113355)
	-- local _, cardLink = GetItemInfo(113346)
	local cardLink = tblCardLinks[idCurrentCard]
	SetOverrideBindingItem(self, true, "MOUSEWHEELUP",   cardLink) -- card of omens
	SetOverrideBindingItem(self, true, "MOUSEWHEELDOWN", cardLink) -- card of omens

	-- SetOverrideBindingItem(self, true, "MOUSEWHEELUP",   "Blood Card") -- card of omens
	-- SetOverrideBindingItem(self, true, "MOUSEWHEELDOWN", "Blood Card") -- card of omens


--	SetOverrideBindingItem(self, true, "MOUSEWHEELUP",   "113355") -- card of omens
--	SetOverrideBindingItem(self, true, "MOUSEWHEELDOWN", 113355) -- card of omens
--	SetOverrideBindingItem(self, true, "MOUSEWHEELUP",   "Card of Omens") -- card of omens
--	SetOverrideBindingItem(self, true, "MOUSEWHEELDOWN", "Card of Omens") -- card of omens
--	SetOverrideBindingMacro(self, true, "MOUSEWHEELUP",   "z-mill") -- card of omens
--	SetOverrideBindingMacro(self, true, "MOUSEWHEELDOWN", "z-mill") -- card of omens
	dbgprint.spam("taking mouse wheel binding")
end

local function frm_Leave(self)
	-- self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	-- self:UnregisterEvent("CHAT_MSG_LOOT")
	ClearOverrideBindings(self)
	dbgprint.spam("releasing mouse wheel binding")
end

function SpinToWin:MakeButtons()

	dbgprint.config("making buttons")
	
	local btn = CreateFrame("Button", "SpinToWinButton", UIParent, "SecureActionButtonTemplate");
	self.btn = btn
	btn:SetAttribute("type", "macro");
	btn:SetAttribute("macrotext", "button pressed")
	btn:SetScript("PreClick",  btn_PreClick);
	-- btn:SetScript("PostClick", btn_PostClick);
	

	local frm = CreateFrame("Frame", "SpinToWinMainFrame", UIParent, "BasicFrameTemplate") --, "SecureActionButtonTemplate")
	self.frm = frm

	frm:SetScript("OnEnter", frm_Enter)
	frm:SetScript("OnLeave", frm_Leave)

	frm:SetPoint("LEFT", UIParent, "LEFT", 100, 0)
	frm:SetSize(300, 150)
	frm:EnableMouse(true)
	frm:SetMovable(true)
	frm:SetResizable(true)
	frm:SetUserPlaced(true)

	--local frmTitle = frm:CreateTitleRegion()
	--self.frmTitle = frmTitle
	--frmTitle:SetAllPoints(frm.TopBorder)
	frm:SetMovable(true)
	frm:RegisterForDrag('LeftButton')
	frm:SetScript('OnDragStart', frm.StartMoving) 
	frm:SetScript('OnDragStop', frm.StopMovingOrSizing)
	
	-- dbgprint.test("resizable", frm:IsResizable(), "movable", frm:IsMovable(), "user placed", frm:IsUserPlaced())

	local function MakeLabel(hjust, vjust, text)
		local newLabel = frm:CreateFontString(nil, "ARTWORK", "GameFontNormal")
		newLabel:SetJustifyH(hjust)
		newLabel:SetJustifyV(vjust)
		newLabel:SetText(text)
		return newLabel
	end
	
	frm.HistoryLabel = MakeLabel("RIGHT", "TOP", L["lifetime"])
	frm.SessionLabel = MakeLabel("RIGHT", "TOP", L["session"])

	-- init flips and gold to have a values of at least 1000 for purposes of setting widths
	-- frm.HistoryMoney = MakeLabel("RIGHT", "TOP", MyGetCoinTextureString( max(10000000, dbglob.totalMoney) ))
	-- frm.HistoryFlips = MakeLabel("CENTER", "TOP", string.format("%i", max(1000, dbglob.totalFlips)))
	frm.HistoryMoney = MakeLabel("RIGHT", "TOP", MyGetCoinTextureString(dbglob.totalMoney))
	frm.HistoryFlips = MakeLabel("CENTER", "TOP", string.format("%i", dbglob.totalFlips))

	frm.SessionMoney = MakeLabel("RIGHT", "TOP", MyGetCoinTextureString(cMoney))
	frm.SessionFlips = MakeLabel("CENTER", "TOP", string.format("%i", cFlips))
	

	local height = frm.HistoryMoney:GetHeight()
	local widthLabels = 10 + max(frm.HistoryLabel:GetStringWidth(), frm.SessionLabel:GetStringWidth())
	-- print("label width", widthLabels)

	frm.HistoryLabel:SetWidth(widthLabels)
	frm.HistoryLabel:SetHeight(height) --, 5, 5)
	frm.HistoryFlips:SetHeight(height) --, 5, 5)

	frm.SessionLabel:SetWidth(widthLabels)
	frm.SessionLabel:SetHeight(height) --, 5, 5)
	frm.SessionFlips:SetHeight(height) --, 5, 5)

	
	frm.HistoryLabel:SetPoint("BOTTOMLEFT", frm, "LEFT", 10, 5)
	frm.HistoryFlips:SetPoint("BOTTOMLEFT", frm.HistoryLabel, "BOTTOMRIGHT", 10, 0)
	frm.HistoryMoney:SetPoint("BOTTOMLEFT", frm, "CENTER", 0, 5)


	frm.SessionLabel:SetPoint("TOPLEFT",  frm.HistoryLabel, "BOTTOMLEFT",  0, -10)
	frm.SessionLabel:SetPoint("TOPRIGHT", frm.HistoryLabel, "BOTTOMRIGHT", 0, -10)
	frm.SessionFlips:SetPoint("TOPLEFT",  frm.HistoryFlips, "BOTTOMLEFT",  0, -10)
	frm.SessionFlips:SetPoint("TOPRIGHT", frm.HistoryFlips, "BOTTOMRIGHT", 0, -10)
	frm.SessionMoney:SetPoint("TOPLEFT",  frm.HistoryMoney, "BOTTOMLEFT",  0, -10)
	frm.SessionMoney:SetPoint("TOPRIGHT", frm.HistoryMoney, "BOTTOMRIGHT", 0, -10)


	
	-- frm.HistoryFlips:SetPoint("TOPRIGHT",    frm.HistoryMoney, "TOPLEFT") --, 5, 5)
	-- frm.SessionFlips:SetPoint("BOTTOMRIGHT", frm.SessionMoney, "BOTTOMLEFT") --, 5, 5)
	
	
	-- dbgprint.test(frm.Session:GetStringHeight(), frm.Session:GetStringWidth())
	-- dbgprint.test("flips", frm.SessionFlips:GetRect())
	-- dbgprint.test("money", frm.SessionMoney:GetRect())
	
	-- dbgprint.test(frm.HistoryFlips:GetJustifyV(), frm.HistoryMoney:GetJustifyV())
	-- dbgprint.test(frm.SessionFlips:GetJustifyV(), frm.SessionMoney:GetJustifyV())
	
	
	-- frm:EnableMouseWheel(true)

	-- frm:SetNormalTexture("Interface\\EncounterJournal\\UI-EncounterJournalTextures")
	-- frm:GetNormalTexture():SetTexCoord(0.25585938, 0.37890625, 0.90332031, 0.95898438)	-- UI-EJ-Tab-Selected

	-- frm:Disable()

end


function SpinToWin:Open()
	if not self.frm then
		self:MakeButtons()
	end

	self.frm:Show()
	-- self.frm:Enable()
end

function SpinToWin:Close()
	if not self.frm then return end

	-- self.frm:Disable()
	self.frm:Hide()
end

-- Called when the addon is enabled
function SpinToWin:OnEnable()
	self:LoadSavedSettings()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpinToWin", SpinToWin.GetUIOptions, { "SpinToWin", "s2w" } )
	
	self:SetDebugLevel(DL[self.db.global.output])

	for _,id in pairs(tblFlippableCards) do
		local _ = tblCardLinks[id]	-- get link data into cache so its ready when we need it
		local t = {}
		for k,v in pairs(tblCardValue) do t[k] = 0 end
		tblFlipsSession[id] = t
	end

	tblFlipsHistory = self.db.global.tblFlips

--	self:SecureHookScript(BonusRollFrame, "OnShow", "WaitForBonusLoot")
--	self:SecureHookScript(BonusRollFrame, "OnHide", "WaitFinished")

	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:RegisterEvent("CHAT_MSG_LOOT")
	-- self:RegisterEvent("ITEM_PUSH")
	
	strItemPush = MakePatternFromFormat(_G.LOOT_ITEM_PUSHED_SELF)
end

-- Called when the addon is disabled
function SpinToWin:OnDisable()
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:UnregisterEvent("CHAT_MSG_LOOT")
	-- self:UnregisterEvent("ITEM_PUSH")

--	self:UnhookAll()
end

function SpinToWin:LoadSavedSettings()
	-- saved settings and default values
	local defaults = {
		global = {
			output = "test",
			reflipCap  = 0,
			totalFlips = 0,
			totalMoney = 0,
			tblFlips = {},

		},
	}

	self.db = LibStub("AceDB-3.0"):New("SpinToWinDB", defaults, true) --, _G.DEFAULT)

	-- we should be able to migrate saved data formats immediately
	local DBVERSION = 1
	dbglob = self.db.global

	dbglob.version = dbglob.version or 0

	if dbglob.version == DBVERSION then
		return
	end

	if dbglob.version < 1 then
		for _,id in pairs(tblFlippableCards) do
			local t = {}
			for k,v in pairs(tblCardValue) do t[k] = 0 end
			dbglob.tblFlips[id] = t
		end
	end

	dbglob.version = DBVERSION
end

function SpinToWin.GetUIOptions(uiType, uiName, appName)
--	print("GetUIOptions   ", uiType)
	
	if (not SpinToWin.options) then
		SpinToWin.options = {}
		
		-- start options.commands --- commands executable from command line
		SpinToWin.options.commands = {
			type = 'group',
			name = "SpinToWin",
			handler = SpinToWin,
			args = {
--[===[
				Debug = {
					hidden = function(info) return SpinToWin.db.global.iDebugLevel == 0 end,
					order = 0,
					type = "input",
					name = "debug",
					desc = "Sets verbosity of debug output",
					get = function(info) return SpinToWin.db.global.iDebugLevel end,
					set = "Debug",
				},
--]===]
				[ L["Open"] ] = {
					order = 10,
					type = "execute",
					name = L["Open"],
					-- desc = L["Decouples the addon from the shared keybind"],
					func = "Open",
				},
				[ L["Close"] ] = {
					order = 20,
					type = "execute",
					name = L["Close"],
					-- desc = L["Treat targeted npc as interactable from now on"],
					func = "Close",
				},
			},
		}  -- end options.commands --- commands executable from command line
	end

	return SpinToWin.options.commands
end

